"""

"""
import napari as na
from napari_animation import Animation


def animate_1axis(viewer: na.Viewer, start_frame: int = 0, end_frame: int = None, axis: int = 0):
    """
    Whatever is currently displayed in viewer gets turned into a napari animation frame-by-frame.
    Use axis to specify temporal dimension.

    Parameters
    ----------
    viewer : na.Viewer
    start_frame : int
    end_frame : int
        Not included.
    axis : int

    Returns
    -------
    animation : Animation
    """
    current_step = viewer.dims.ndim * [0,]

    if end_frame is None:
        end_frame = int(viewer.dims.range[axis][1])

    animation = Animation(viewer)
    for frame in range(start_frame, end_frame): # end_frame is included
        current_step[axis] = frame
        viewer.dims.current_step = tuple(current_step)
        animation.capture_keyframe(steps=1)

    return animation

def save_animation(animation: Animation,
                   save_path: str,
                   canvas_only: bool = True,
                   fps: float = 24,
                   quality: float = 5):
    """
    Render and save napari_animation.Animation

    Parameters
    ----------
    animation : Animation
    save_path : str
        Where animation gets saved. Recommended: include .avi extension.
    canvas_only : bool
    fps : float
    quality : float
    """
    animation.animate(
        save_path,
        canvas_only = canvas_only,
        fps = fps,
        quality = quality
    )


"""
Methods for converting images to other formats.
"""
from glob import glob
from os import makedirs
from os.path import join, exists
from pathlib import Path

import h5py
import hdf5storage
import nd2
import numpy as np
import pandas as pd
from scipy.io import savemat
import tifffile

def cellpose_to_mat(file_cp, file_mat = None, ext_cp = '_seg.npy', ext_mat = '_PhaseMask.mat', folder_dest = None):
    """
    Converts .npy file in cellpose format to .mat file that is readable by SMALL-LABS.

    Parameters
    ----------
    file_cp : str
        .npy file created by cellpose
    file_mat : str, optional
        Name of .mat to be created. If None, has same name as cellpose file with '_PhaseMask.mat' appended.
    ext_cp : str, optional
        Cellpose extension. Gets chopped off before saving .mat.
    ext_mat : str, optional
        Extension. Appended to end of file name for saving. Default: '_PhaseMask.mat'
    folder_dest : str, optional
        Destination folder. If None, .mat saved in same folder as cellpose file.

    Returns
    -------
    None
    """
    if folder_dest is None:
        folder_dest = Path(file_cp).parent
    elif not exists(folder_dest):
        makedirs(folder_dest)

    if file_mat is None:
        stem_mat = Path(file_cp).split(ext_cp)[0]
        file_mat = join(folder_dest, stem_mat + ext_mat)

    masks = np.load(file_cp, allow_pickle=True).item()['masks']
    savemat(file_mat, {'PhaseMask': masks})

def cellpose_to_mat_batch(folder_cp, pattern_cp = '*_seg.npy', ext_cp = '_seg.npy', ext_mat = '_PhaseMask.mat'):
    """
    Converts cellpose style .npy files to .mat files readable by SMALL-LABS.

    Parameters
    ----------
    folder_cp : str
        Folder containing cellpose .npy files.
    patter_cp : str, optional
        Pattern used to find cellpose output files.
    ext_cp : str, optional
        Part of cellpose file names to be removed.
    ext_mat : str, optional
        Added to end of filename (after ext_cp removed).

    Returns
    -------
    None
    """
    files_cp = glob(join(folder_cp, pattern_cp))

    [cellpose_to_mat(file, ext_cp=ext_cp, ext_mat=ext_mat) for file in files_cp]

def cellpose_to_labels(filename):
    """
    Get labels (2d numpy array) generated from _seg.npy file generated by Cellpose.
    
    Parameters
    ----------
    filename : str

    Returns
    -------
    2d array type int
    """
    return np.load(filename, allow_pickle=True).item()['masks']

def df_to_sl(df, col_mapper = None):
    """
    Convert pandas DataFrame to SMALL-LABS .mat format.

    Parameters
    ----------
    df : pd.DataFrame
    col_mapper : dict
        Keys are columns in df, values are new names in SMALL-LABS format. Used if, for example,
        columns in df, 'x' and 'y', need to be re-labeled 'col' and 'row', respectively, 
        for use in NOBIAS.
        Example:
            col_mapper = {'x': 'col', 'y': 'row', 'n': 'track_id', 'rois': 'roinum'}
        
        Columns 'frame', 'row', 'col', 'track_id', and 'roinum' are required to generate tracks.

    Returns
    -------
    data_sl : dict
        Orgainized like SMALL-LABS data structure. 
        First level keys: 'fits', 'tracks'.
        Second level: 'fits' additional keys matching columns in df after renaming.
    """
    if col_mapper is not None:
        data = df.rename(columns=col_mapper) # makes a copy of df
    else:
        data = df

    data_sl = {}

    fits_dict = {c: data[c].values for c in data.columns} # may need to play with axes ordering
    data_sl['fits'] = fits_dict

    # to have 'frame', 'row', 'col', 'track_id' columns to generate tracks
    if set(('frame', 'row', 'col', 'track_id')).issubset(data.columns):
        tracks = np.zeros([len(data), 6])
        tracks[:,0] = data['frame'].values
        tracks[:,1] = data['row'].values
        tracks[:,2] = data['col'].values
        tracks[:,3] = data['track_id'].values

        if 'roinum' in data.columns: tracks[:,4] = data['roinum'].values
        else: tracks[:,4] = np.ones(len(data))

        if 'molid' in data.columns: tracks[:,5] = data['molid'].values
        else: tracks[:,5] = data.index.values

        data_sl['tracks'] = tracks

    return data_sl

def df_to_so(data_df, frame_col='frame', coord_cols=['x', 'y'], track_col='track_id', pixel_size=1, t_frame=0.04):
    """
    Convert tracks in pandas DataFrame to numpy array format expected by Spot-On.

    Parameters
    ----------
    data_df : pd.DataFrame
        Tracking data.
    frame_col : str
        Column containing frame number.
    coord_cols : List[str]
        Column names for 'x' and 'y' coordinates, respectively.
    track_col : str
    pixel_size : float
        For converting to um.
    t_frame : float
        Time between frames in seconds.

    Returns
    -------
    so_tracks : np.array, dtype=[('xy', 'O'), ('TimeStamp', 'O'), ('Frame', 'O')]
        Spot-On formatted tracks
    """
    frames = data_df[frame_col].values
    xy = data_df[coord_cols].values * pixel_size
    frames = (data_df[frame_col].values).astype('int')
    times = frames * t_frame
    track_ids = data_df[track_col].values
    unique_track_ids = np.unique(track_ids)
    so_tracks = [(xy[track_ids==ti], times[track_ids==ti], frames[track_ids==ti]) for ti in unique_track_ids]
    so_tracks = np.array(so_tracks, dtype=[('xy', 'O'), ('TimeStamp', 'O'), ('Frame', 'O')])

    return so_tracks

def nd2_to_npy(file_nd2, file_npy = None, ext = '.npy', folder_dest = None):
    """
    Converts an nd2 file (Nikon) to a .npy (numpy array).
    
    Parameters
    ----------
    file_nd2 : str
        Absolute path to nd2 file.
    file_npy : str, optional
        Name of .tif to be created. If None, has same name as nd2 plus ext.
    ext : str, optional
        Extension. Appended to end of file name for saving. Default: '.npy'
    folder_dest : str, optional
        Destination folder. If None, .npy saved in same folder as nd2.

    Returns
    -------
    None
    """
    if folder_dest is None:
        folder_dest = Path(file_nd2).parent
    elif not exists(folder_dest):
        makedirs(folder_dest)

    if file_npy is None:
        file_npy = join(folder_dest, Path(file_nd2).stem + ext)

    image = nd2.imread(file_nd2)
    np.save(file_npy, arr=image)

def nd2_to_npy_batch(folder_nd2, pattern_nd2 = '*.nd2', ext = '.npy', folder_dest = None):
    """
    Converts .nd2 files to .npy files (numpy arrays).

    Parameters
    ----------
    folder_nd2 : str
        Folder containing .nd2 files.
    patter_nd2 : str, optional
        Pattern used to find .nd2 files. Default: '*.nd2'.
    ext : str, optional
        Extension. Appended to end of file name for saving. Default: '.npy'.
    folder_dest : str, optional
        Destination folder. If None, .npy saved in same folder as nd2.

    Returns
    -------
    None
    """
    files_nd2 = glob(join(folder_nd2, pattern_nd2))

    for file in files_nd2:
        nd2_to_npy(file, ext=ext, folder_dest=folder_dest)

def nd2_to_tif(file_nd2, file_tif = None, ext = '.tif', folder_dest = None):
    """
    Converts an nd2 file (Nikon) to a tif.

    Parameters
    ----------
    file_nd2 : str
        Absolute path to nd2 file.
    file_tif : str, optional
        Name of .tif to be created. If None, has same name as nd2 with .tif extension.
    ext : str, optional
        Extension. Appended to end of file name for saving. Default: '.tif'
    folder_dest : str, optional
        Destination folder. If None, tif saved in same folder as nd2

    Returns
    -------
    None
    """
    if folder_dest is None:
        folder_dest = Path(file_nd2).parent
    elif not exists(folder_dest):
        makedirs(folder_dest)

    if file_tif is None:
        file_tif = join(folder_dest, Path(file_nd2).stem + ext)

    image = nd2.imread(file_nd2)
    tifffile.imsave(file_tif, data=image)

def nd2_to_tif_batch(folder_nd2, pattern_nd2 = '*.nd2', ext = '.tif', folder_dest = None):
    """
    Convert .nd2 files to .tif files.

    Parameters
    ----------
    folder_nd2 : str
        Path to folder containing .nd2 files.
    pattern_nd2 : str, optional
        Pattern used to find .nd2 file. Default: '*.nd2'.
    ext : str, optional
        String append to file name when saving .tif.
    folder_dest : str, optional
        Folder where .tif files saved. If none is provided, .tif files are saved to same folder
        as .nd2 files.

    Returns
    -------
    None
    """
    files_nd2 = glob(join(folder_nd2, pattern_nd2))

    for file in files_nd2:
        nd2_to_tif(file, ext=ext, folder_dest=folder_dest)

def save_sl_fits(data_sl, file_sl):
    """
    Save dict structured SMALL-LABS-like as .mat. Wrapper around hdf5storage.savemat(...).

    Parameters
    ----------
    data_sl : dict
    file_sl : str
        name of file where data_sl is to be save. Should end with '.mat'.

    Returns
    -------
    None
    """
    hdf5storage.savemat(file_name = file_sl, mdict = data_sl, format = '7.3')

def slfile_to_csv(file_sl, file_csv = None, ext = '.csv', folder_dest = None):
    """
    Reads SMALL-LABS .mat file and saves as .csv file.

    Parameters
    ----------
    file_sl : str
        .mat filename. Full path.
    file_csv : str, optional
        Name of new file where data is to be stored as .csv. Full path.
    ext : str, optional
        Bit of text added to end of new .csv name. Default is '.csv'.
    folder_dest : str, optional
        Destination folder. If none provided, .csv saved in origin folder.

    Returns
    -------
    None
    """
    if folder_dest is None:
        folder_dest = Path(file_sl).parent
    elif not exists(folder_dest):
        makedirs(folder_dest)

    if file_csv is None:
        file_csv = join(folder_dest, Path(file_sl).stem + ext)

    obj_sl = h5py.File(file_sl)
    df = sl_to_df(obj_sl)

    df.to_csv(file_csv)

def slfile_to_df(file_sl: str):
    """
    Read SMALL-LABS file and convert to DataFrame.

    Parameters
    ----------
    file_sl : str

    Returns
    -------
    df : pd.DataFrame
    """
    obj_sl = h5py.File(file_sl)
    df = sl_to_df(obj_sl)

    return df


def slfile_to_csv_batch(folder_sl, pattern_sl = '*_fits.mat', ext = '.csv', folder_dest = None):
    """
    Converts SMALL-LABS files to csv.
    
    Parameters
    ----------
    folder_sl : str
        Folder containing .mat files to be converted to .csv.
    pattern_sl : str, optional
        Pattern used to identify files to be converted.
    ext : str, optional
        Bit of text added to end of new .csv names. Default is '.csv'.
    folder_dest : str, optional
        Destination folder. If none provided, .csv saved in origin folder.

    Returns
    -------
    None
    """
    files_sl = glob(join(folder_sl, pattern_sl))

    [slfile_to_csv(file, ext=ext, folder_dest=folder_dest) for file in files_sl]

def sl_to_df(sl_object):
    """
    Convert SMALL-LABS object read from .mat to pandas DataFrame.
    
    Parameters
    ----------
    sl_object : SMALL-LABS data object, a .mat file read using h5py

    Returns
    -------
    pd.DataFrame
    """
    if 'fits' in sl_object.keys():
        df = pd.DataFrame(
            data={
                key: sl_object['fits'][key][0,:] for key in sl_object['fits'].keys() if key not in ('rowCI', 'colCI')
                }
            )
        if 'rowCI' in sl_object['fits'].keys():
            if len(sl_object['fits']['rowCI'].dtype) == 2:
                rowCI = np.array([[re, im] for (re, im) in sl_object['fits']['rowCI'][0]])
                df['rowCI_re'] = rowCI[:,0]
                df['rowCI_im'] = rowCI[:,1]
            elif len(sl_object['fits']['rowCI'].dtype) == 1:
                df['rowCI'] = sl_object['fits']['rowCI'][0,:]
        if 'colCI' in sl_object['fits'].keys():
            if len(sl_object['fits']['colCI'].dtype) == 2:
                colCI = np.array([[re, im] for (re, im) in sl_object['fits']['colCI'][0]])
                df['colCI_re'] = colCI[:,0]
                df['colCI_im'] = colCI[:,1]
            elif len(sl_object['fits']['colCI'].dtype) == 1:
                df['colCI'] = sl_object['fits']['colCI'][0,:]
    elif 'guesses' in sl_object.keys():
        df = pd.DataFrame(data={'frame': sl_object['guesses'][0,:],
                                'row': sl_object['guesses'][1,:],
                                'col': sl_object['guesses'][2,:]})
        if 'roinum' in sl_object.keys():
            df['roinum'] = sl_object['roinum'][0,:]

    df.loc[:,'frame'] -= 1
    df.loc[:,'row'] -= 1
    df.loc[:,'col'] -= 1

    if 'trk_filt' in sl_object.keys():
        df['trk_filt'] = sl_object['trk_filt'][0,:]

    if 'tracks' in sl_object.keys():
        # tracks_df = pd.DataFrame(data=sl_object['tracks'][:].T,
        #                          columns=('frame', 'row', 'col', 'track_id', 'roinum', 'molid'))
        tracks_df = pd.DataFrame(data={'track_id': sl_object['tracks'][3,:], 'molid': sl_object['tracks'][5,:]})
        df = pd.merge(df, tracks_df, how='left')
        # df = df.merge(tracks_df, how='left')
        df['tracked'] = ~np.isnan(df['track_id'])
        
        # df['tracked'] = np.isin(df['molid'], sl_object['tracks'][5,:])
        # df['track_id'] = np.nan
        # df.loc[df['tracked']==True, 'track_id'] = sl_object['tracks'][3,:] # wrong


        # if 'roinum' in df.columns:
        #     track_id_max = df['track_id'].max()
        #     df['track_id_unique'] = df['track_id'] + (df['roinum'] - 1)*track_id_max

    return df

def sl_to_df_batch(objs_sl: list):
    """
    Converst list of SMALL-LABS objects to list of pandas DataFrames.

    Parameters
    ----------
    objs_sl : list (or other iterable)
        SMALL-LABS objects to be converted to pandas DataFrames.

    Returns
    -------
    list
        Each element is pd.DataFrame
    """
    return [sl_to_df(o) for o in objs_sl]

def sl_to_so(data_sl, pixel_size=0.049, t_frame=0.04):
    """
    Convert tracks in SMALL-LABS-style dict to numpy array format expected by Spot-On.

    Parameters
    ----------
    data_sl : dict
    pixel_size : float
        Spot-On expects coordinates in be in micron, so um/pixel required to convert.
    t_frame : float
        Time between frames in seconds.

    Returns
    -------
    so_tracks : np.array, dtype=[('xy', 'O'), ('TimeStamp', 'O'), ('Frame', 'O')]
        Spot-On formatted tracks
    """
    if 'tracks' in data_sl.keys():
        data_np = data_sl['tracks'][:].T # numpy array
        if data_np.ndim == 2: # False if SMALL-LABS didn't find any tracks
            xy = data_np[:,1:3] * pixel_size
            frames = (data_np[:,:1].T).astype('int')
            times = frames * t_frame
            track_ids = data_np[:,3]
            unique_track_ids = np.unique(track_ids)
            so_tracks = [(xy[track_ids==ti], times[track_ids==ti], frames[track_ids==ti]) for ti in unique_track_ids]
            so_tracks = np.array(so_tracks, dtype=[('xy', 'O'), ('TimeStamp', 'O'), ('Frame', 'O')])
        else:
            so_tracks = np.array([])
    else:
        so_tracks = np.array([])

    return so_tracks


"""

"""
import matplotlib as mpl
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
import matplotlib_scalebar.scalebar as sb
import numpy as np
import pandas as pd
from skimage.measure import find_contours, regionprops


default_scalebar_props = {
    'dx': 0.049,
    'units': 'um',
    'fixed_value': 4,
    'scale_loc': 'none',
    'location': 'lower left',
    'frameon': False
}

transparent_rgba = mpl.colors.to_rgba('xkcd:white', alpha=0)
black = mpl.colors.to_rgba('xkcd:black', alpha=1)
hotmagenta_tr = mpl.colors.to_rgba('xkcd:hot magenta', alpha=0)
hotmagenta_op = mpl.colors.to_rgba('xkcd:hot magenta', alpha=1)

transparent2hotmagenta = mpl.colors.LinearSegmentedColormap.from_list('transparent2hotmagenta',
                                                                   [hotmagenta_tr, hotmagenta_op],
                                                                   512)

black2hotmagenta = mpl.colors.LinearSegmentedColormap.from_list('black2hotmagenta',
                                                                   [black, hotmagenta_op],
                                                                   512)

  
def crop_to_labels(ax, labels, crop_buffer=5, scale=1):
    """
    Set limits of ax so that only bounding box of labels is shown with crop_buffer number of
    pixels added to border.

    Parameters
    ----------
    ax : matplotlib axis handle
    labels : np.array, dtype=int
    crop_buffer : int
        Extra pixels to leave around bounding box. Always in pixels even if scale is set.
    scale : float
        For converting pixel coordinates to other units.

    Returns
    -------
    None
    """
    nrows_init, ncols_init = labels.shape

    labels_bool = (labels > 0).astype('int')
    rowmin, colmin, rowmax, colmax = regionprops(labels_bool)[0]['bbox']

    xmin = np.max([-0.5, colmin-crop_buffer-0.5]) * scale
    xmax = np.min([colmax+crop_buffer+0.5, ncols_init-0.5]) * scale
    ymin = np.max([-0.5, rowmin-crop_buffer-0.5]) * scale
    ymax = np.min([rowmax+crop_buffer+0.5, nrows_init-0.5]) * scale
    
    ax.set_xlim(left=xmin, right=xmax)
    ax.set_ylim(bottom=ymax, top=ymin)


def figs_to_pdf(figures, save_name):
    """
    Save series of figures to pdf where each figure is on a single page.

    Parameters
    ----------
    figures : List[matplotlib figure objects]
        List of figure handles.
    savename : str, Path

    Returns
    -------
    None
    """
    image_pdf = PdfPages(save_name)
    for fig in figures:
        image_pdf.savefig(fig)
    image_pdf.close()


def gauss_kernel(x, sigma):
    """
    Helper for smoothening polygons.
    """
    gk = np.exp(-x**2 / (2 * sigma**2))
    gk[x==0] = 0
    gk /= gk.sum()
    return gk


def labels_to_contours(labels, level=0.5, smooth=False, lam=0.39, mu=-0.4, N=500, sigma=1.5):
    """

    Parameters
    ----------
    labels : np.array(dtype=int)
    level : float, (0, 1)

    Returns
    -------
    contours : list[np.array(shape=(N,2))]
        Length of contours matches number of rois
        Columns of each array contain rows and columns, respectively.
    """
    n_labels = labels.max()
    contours = []
    for l in range(1, n_labels+1):
        label = labels == l
        contour = find_contours(label, level=level)
        contours.append(*contour)

    if smooth == True:
        contours = [smooth_polygon(contour, 
                                   lam=lam, 
                                   mu=mu, 
                                   N=N, 
                                   sigma=sigma) for contour in contours]
        
    return contours


def plot_locs_gaussian():
    """
    Plots localizations with each represented by a Gaussian distribtuion.

    Parameters
    ----------
    

    Returns
    -------
    fig
    ax
    """
    pass


def plot_locs_scatter(
        locs_df,
        coord_cols=('x', 'y'),
        scale = 1,
        labels = None,
        image = None,
        crop = False,
        figure_props = {},
        marker_props = {},
        scalebar_props = None,
        smooth = True,
        smooth_parameters = {}
        ):
    """
    Plot single-molecule localization data as a scatter plot.
    
    Parameters
    ----------
    locs_df : pd.DataFrame
    coord_cols : 2-tuple of strings, default ('x', 'y')
        Specify columns of locs_df containing coordinates.
        Set to ('col', 'row') for data originating from SMALL-LABS.
    scale : float
        Factor to convert units of coord_cols to pixels
    labels : np.ndarray
        2d integer array labeling regions of interest.
        Typically the output of segmentation, e.g. cellpose.
    image : np.ndarray
        2d array to overlay localization scatter on.
        Typically, some kind reference image, e.g. phase contrast.
        If None provided, plot on black background by default.
    crop : bool, default False
        If True, will crop to only display area where labels > 0.
    figure_props : dict
        Inputs to plt.subplots.
    marker_props : dict
        kwargs to matplotlib.pyplot.scatter
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.scatter.html
    scalebar_props : dict
        Inputs to add_scalebar. If None provided (default), no scalebar is drawn.
        
    Returns
    -------
    fig, ax
    """
    xcol, ycol = coord_cols

    if labels is not None:
        contours = labels_to_contours(labels, smooth=smooth, **smooth_parameters)
    else:
        contours = []

    if image is None and labels is None:
        colmax = int(np.max(locs_df[xcol]*scale)) + 1
        rowmax = int(np.max(locs_df[ycol]*scale)) + 1
        image = np.zeros([rowmax, colmax])
    elif image is None:
        image = np.zeros(labels.shape)

    fig, ax = plt.subplots(1, 1, **figure_props)
        
    ax.imshow(image, cmap='binary_r')
    
    for contour in contours:
        ax.plot(contour[:,1], contour[:,0], lw=1, alpha=0.5, color='xkcd:gray')

    ax.scatter(locs_df[xcol]*scale, locs_df[ycol]*scale, **marker_props)
            
    if crop == True and labels is not None:
        crop_to_labels(ax, labels)
    else:
        ax.set_xlim(left=-0.5, right=image.shape[1] - 0.5)
        ax.set_ylim(bottom=image.shape[0] - 0.5, top=-0.5)

    if scalebar_props is not None:
        scalebar = sb.ScaleBar(**scalebar_props)
        ax.add_artist(scalebar)

    ax.axis('off')

    return fig, ax


def plot_overlay(
        image_bgd: np.ndarray,
        image_fgd: np.ndarray,
        vmin_bgd: float = None,
        vmax_bgd: float = None,
        vmin_fgd: float = None,
        vmax_fgd: float = None,
        cmap_bgd: str = "binary_r",
        cmap_fgd_solo: mpl.colors.LinearSegmentedColormap = black2hotmagenta,
        cmap_fgd_overlay: mpl.colors.LinearSegmentedColormap = transparent2hotmagenta,
        scalebar_props: dict = default_scalebar_props
        ):
    """
    Creates figure with three axes for displaying two images and an overlay:
    1) Background image, e.g. phase contrast.
    2) Foreground image, e.g. bulk fluorescence.
    3) Overlay of foreground image on background image.

    Parameters
    ----------
    image_bgd : np.ndarray
        2D array for background image.
    image_fgd : np.ndarray
        2D array for foreground image.
    vmin_bgd : float
        Lower limit on background image - for setting contrast. Default is image_bgd.min().
    vmax_bgd : float
        Upper limit on background image - for setting contrast. Default is image_bgd.max().
    vmin_fgd : float
        Lower limit on foreground image - for setting contrast. Default is image_fgd.min().
    vmax_fgd : float
        Upper limit on foreground image - for setting contrast. Default is image_fgd.max().
    cmap_bgd : str, default "binary_r"
        Colormap for background image.
    cmap_fgd_solo : mpl.colors.LinearSegmentedColormap
        Colormap for foreground image in middle axes. Default is hot magenta against black.
    cmap_fgd_overlay : mpl.colors.LinearSegmentedColormap
        Colormap for foreground image in overlay. Default is hot magenta against transparent.
    scalebar_props : dict

    Returns
    -------
    fig : mpl.figure.Figure
    ax : mpl.axes.Axes
    """
    scalebar = sb.ScaleBar(**scalebar_props)

    fig, ax = plt.subplots(1, 3, figsize=(12,6))

    ax[0].imshow(image_bgd,
                 cmap=cmap_bgd,
                 vmin=vmin_bgd,
                 vmax=vmax_bgd)
    ax[0].add_artist(scalebar)
    ax[0].axis('off')

    ax[1].imshow(image_fgd,
                cmap=cmap_fgd_solo,
                vmin=vmin_fgd,
                vmax=vmax_fgd)
    ax[1].axis('off')

    ax[2].imshow(image_bgd,
                 cmap=cmap_bgd,
                 vmin=vmin_bgd,
                 vmax=vmax_bgd)
    ax[2].imshow(image_fgd,
                cmap=cmap_fgd_overlay,
                vmin=vmin_fgd,
                vmax=vmax_fgd)
    ax[2].axis('off')

    plt.tight_layout()

    return fig, ax


def plot_tracks(
        locs_df,
        track_data = None,
        coord_cols=('x', 'y'),
        track_col = 'track_id',
        color_col = None,
        track_data_bins = np.array([[-np.inf, np.inf]]),
        track_data_colors = None,
        order = 'forward',
        scale = 1,
        labels = None,
        image = None,
        subsample = None,
        line_props = {},
        separate = False,
        crop = False,
        figure_props = {},
        scalebar_props = None
        ):
    """
    Plot single-molecule tracks.

    Parameters
    ----------
    locs : pd.DataFrame
        Localization data. Required columns: 'x', 'y', 'track_id'.
    track_data : pd.DataFrame, optional
        Columns: 'track_id' + other values calculated for each track. Used to filter, choose color.
    track_data_bins : array_like
        Nx2 array where each row defines a half open interval for categorizing 
    track_data_colors : iterable
    order : str
        'forward' or 'reverse'.
        Determines which order to plot binned track data.
    scale : float, default 1
        Used to convert localizations to units of pixels if not already.
    labels : 2d int array
        Segmentation data for cell, nuclei, etc.
    subsample : int, default None
        Number of tracks to plot. If None, all tracks are plotted.
    line_props : dict, optional
    separate : bool, default False
    crop : bool, default False
    figure_props : dict, optional
    scalebar_props : dict, optional

    Returns
    -------
    fig
        Matplotlib figure handle.
    ax
        Matplotlib axis handle(s).
    """
    xcol = coord_cols[0]
    ycol = coord_cols[1]

    if track_data is None:
        track_ids = np.unique(locs_df[track_col])
        track_data = pd.DataFrame(data={track_col: np.unique(track_ids)})

    if subsample is not None:
        idx_sub = np.random.choice(track_data.index, size=subsample, replace=False)
        track_data = track_data.loc[idx_sub].reset_index()

    if labels is not None:
        contours = labels_to_contours(labels)
    else:
        contours = []

    if image is None and labels is None:
        colmax = int(np.max(locs_df[xcol]*scale)) + 1
        rowmax = int(np.max(locs_df[ycol]*scale)) + 1
        image = np.zeros([rowmax, colmax])
    elif image is None:
        image = np.zeros(labels.shape)

    n_bins = track_data_bins.shape[0]

    if track_data_colors is None:
        track_data_colors = ['xkcd:hot pink'] * n_bins
    n_colors = len(track_data_colors)

    if separate == True:
        fig, ax = plt.subplots(n_bins, 1, **figure_props)
        if n_bins == 1: ax = [ax]
        for i_ax, (l, h) in enumerate(track_data_bins):
            
            ax[i_ax].imshow(image, cmap='binary_r')
            ax[i_ax].axis('off')

            for contour in contours:
                ax[i_ax].plot(contour[0][:,1], contour[0][:,0],
                            lw=1,
                            alpha=0.5,
                            color='xkcd:gray') # future: contour props
            
            filt = (track_data[track_col] > l) & (track_data[track_col] <= h)
            for ti in track_data[track_col][filt]:
                loc_filt = locs_df[track_col] == ti
                x = locs_df[xcol][loc_filt]
                y = locs_df[ycol][loc_filt]
                ax[i_ax].plot(x*scale, y*scale, color=track_data_colors[i_ax%n_colors], **line_props)

            if crop == True:
                crop_to_labels(ax[i_ax], labels)

                if scalebar_props is not None:
                    # add_scalebar(ax, scalebar_props)
                    scalebar = sb.ScaleBar(**scalebar_props)
                    ax.add_artist(scalebar)

    elif separate == False:
        fig, ax = plt.subplots(1, 1, **figure_props)
        
        ax.imshow(image, cmap='binary_r')
        ax.axis('off')

        for contour in contours:
            ax.plot(contour[0][:,1], contour[0][:,0], lw=1, alpha=0.5, color='xkcd:gray')

        if order == 'forward':
            i_bin = np.arange(n_bins, dtype=int)
        elif order == 'reverse':
            i_bin = np.arange(n_bins-1, -1, -1, dtype=int)

        for i in i_bin:
            l, h = track_data_bins[i]
            filt = (track_data[color_col] > l) & (track_data[color_col] <= h)
            for ti in track_data[track_col][filt]:
                loc_filt = locs_df[track_col] == ti
                x = locs_df[xcol][loc_filt]
                y = locs_df[ycol][loc_filt]
                ax.plot(x*scale, y*scale, color=track_data_colors[i%n_colors], **line_props)
                
        if crop == True:
            crop_to_labels(ax, labels)

        if scalebar_props is not None:
            scalebar = sb.ScaleBar(**scalebar_props)
            ax.add_artist(scalebar)

    return fig, ax


def smooth_polygon(polygon, lam=0.39, mu=-0.4, N=500, sigma=1.5):
    """

    """
    nv = len(polygon) - 1
    I = np.identity(nv)
    W = np.zeros(I.shape)
    
    x = np.arange(-nv//2, nv//2)
    gk = gauss_kernel(x, sigma)
    gk = np.roll(gk, -nv//2)
    
    for v in range(nv):
        W[v,:] = gk
        gk = np.roll(gk, 1)

    K = I - W
    smooth_operator = np.linalg.matrix_power(np.matmul((I - mu*K), (I - lam*K)), N)
    polygon_smooth = np.matmul(smooth_operator, polygon[:-1]) # ignore second end
    polygon_smooth = np.concatenate([polygon_smooth,
                                    [polygon_smooth[0]]]) # make ends match
    
    return polygon_smooth


"""
Operations for manipulating pandas DataFrames containing single-molecule localization data.
"""

import numpy as np
import pandas as pd

def filter_by_nlocs(locs_df, min_locs=2, max_locs=np.inf, track_col='track_id'):
    """
    Remove data for tracks with fewer localizations than min_locs and greater than max_locs.

    Parameters
    ----------
    locs_df : pd.DataFrame
    min_locs : int, default 2
    max_locs : int, default np.inf
    track_col : str, default 'track_id'
        Name of column that identifies which track a localization belongs to.

    Returns
    -------
    locs_filtered : pd.DataFrame
    """
    locs_ngroups = locs_df.groupby(track_col)
    locs_filtered = locs_ngroups.filter(lambda x: (x[track_col].count() >= min_locs) & (x[track_col].count() <= max_locs))

    return locs_filtered.reset_index()

def locs_to_steps(locs_df, frame_col='frame', coord_cols=('x', 'y'), track_col='track_id'):
    """
    From locs_df calculate data for individual steps.

    Parameters
    ----------
    locs_df : pd.DataFrame
    frame_col : str, default 'frame'
    coord_cols : tuple, default ('x', 'y')
    track_col : str, default 'track_id'

    Returns
    -------
    steps_df : pd.DataFrame
        DataFrame with columns: frame0, frame1, x0, x1, y0, y1, xmid, ymid, d, gap
        frame0, x0, y0 are start coordinates for step each step.
        frame1, x1, y2 are end coordinates.
        xmid, ymid are the midpoint.
        d is the length.
        gap is the number of frames between the start and end.
    """
    xcol, ycol = coord_cols

    locs_sorted = locs_df.sort_values([track_col, frame_col], axis=0, ascending=True)

    frames = locs_sorted[frame_col].values
    x, y = locs_sorted[xcol].values, locs_sorted[ycol].values
    track_id = locs_sorted[track_col].values

    track_filt = track_id[:-1] == track_id[1:] # to verify 

    steps_df = pd.DataFrame(data={'frame0': frames[:-1][track_filt], 'frame1': frames[1:][track_filt],
                                  'x0': x[:-1][track_filt], 'x1': x[1:][track_filt],
                                  'y0': y[:-1][track_filt], 'y1': y[1:][track_filt]})

    steps_df['xmid'] = (steps_df['x0'] + steps_df['x1']) / 2
    steps_df['ymid'] = (steps_df['y0'] + steps_df['y1']) / 2
    steps_df['d'] = np.sqrt((steps_df['x0'] - steps_df['x1'])**2 + (steps_df['y0'] - steps_df['y1'])**2)
    steps_df['gap'] = steps_df['frame1'] - steps_df['frame0']

    return steps_df

def median_step_size(locs_df, frame_col='frame', coord_cols=['x', 'y'], track_col='track_id'):
    """
    Finds median step size for each unique track in locs_df.

    Parameters
    ----------
    locs_df : pd.DataFrame
    frame_col : str
    coord_cols : iterabale[str]
    track_col : str, default 'track_id'
        Name of column that identifies which track a localization belongs to.

    Returns
    -------
    d_med : np.ndarray
    track_ids : np.ndarray
    """
    track_ids = np.unique(locs_df['track_id'].values)
    frames = locs_df[frame_col].values
    x = locs_df[coord_cols[0]].values
    y = locs_df[coord_cols[1]].values

    d_med = np.zeros(len(track_ids))
    for i, track_id in enumerate(track_ids):
        track_filt = locs_df[track_col].values == track_id
        if track_filt.sum() > 1:
            frames_track = frames[track_filt]
            x_track = x[track_filt]
            y_track = y[track_filt]
            d = np.sqrt((x_track[1:] - x_track[:-1])**2 + (y_track[1:] - y_track[:-1])**2)
            d = d[frames_track[1:] - frames_track[:-1] == 1]
            d_med[i] = np.median(d)
        else:
            d_med[i] = np.nan

    return pd.DataFrame(data={track_col: track_ids, 'd_med': d_med})


import setuptools


with open("readme.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name='biteen_utilities',
    version='0.1.1',
    author='Daniel Foust',
    author_email='djfoust@umich.edu',
    description='Tools for working with single-molecule localization data.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(),
    install_requires=[
        'h5py',
        'hdf5storage',
        'matplotlib',
        'matplotlib_scalebar',
        'napari[all]',
        'napari_animation',
        'nd2',
        'numpy',
        'pandas',
        'scikit-image',
        'scipy',
        'tifffile',
    ],
    classifiers=(
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    )
)